from pathlib import Path


def get_variants_list(root_dir: Path) -> list[str]:
    """Generates a list of variants based on subdir names of:
     [PROJECT_ROOT]/variants

    :return: list of variants
    """
    variants_dir = root_dir / 'variants'
    variants = []

    if variants_dir.is_dir():
        for variant in variants_dir.iterdir():
            if variant.is_dir():
                pass

        variants = [variant.name
                    for variant in variants_dir.iterdir() if variant.is_dir()]

    return variants


import subprocess

import pytest
from pytest_mock.plugin import MockerFixture

from doimbu.sources import git
from doimbu.version import Version


@pytest.mark.parametrize(
    'git_list_tag_cmd_return,expected_version',
    [
        pytest.param(
            b'',  # git_list_tag_cmd_return
            Version(version='0.1.0'),  # expected_version
            id='no tag'),
        pytest.param(
            b'1.0\n',  # git_list_tag_cmd_return
            Version(version='1.0.0'),  # expected_version
            id='one good tag - 1.0'),
        pytest.param(
            b'1.0.1\n2.2.3\n3.1\n',  # git_list_tag_cmd_return
            Version(version='3.1.0'),  # expected_version
            id='one good tag - 3.1.0'),
        pytest.param(
            b'1.0.1\n2.2.3\n3.1\nsomething_else\n',  # git_list_tag_cmd_return
            Version(version='3.1.0'),  # expected_version
            id='multiple tags with latest non version tag - somthing else'),
        pytest.param(
            b'something_else\n',  # git_list_tag_cmd_return
            Version(version='0.1.0'),  # expected_version
            id=' non-version tag - somthing else'),
    ]
)
def test_get_latest_git_version_tag(
        git_list_tag_cmd_return: bytes,
        expected_version: Version,
        mocker: MockerFixture) -> None:
    mocker.patch(
        'doimbu.sources.git.subprocess.run',
        return_value=subprocess.CompletedProcess(
            [], 0, stdout=git_list_tag_cmd_return))

    latest_git_version = git.get_latest_git_version_tag()

    assert latest_git_version == expected_version


import subprocess
from argparse import Namespace
from configparser import ConfigParser
from pathlib import Path

from pydantic import BaseModel, ConfigDict

from doimbu.paths import BUILDER_CONFIG_PATH, EXECUTION_DIR
from doimbu.sources.file_config import get_file_config
from doimbu.sources.dir_structure import get_variants_list


class BuilderConfig(BaseModel):
    version: str
    default_variant: str
    tag_namespace: str | None = None
    tag_repository: str
    build_args: dict[str, str] | None = None
    # Change default behavior flags
    only_default_variant: bool = False
    tag_omit_latest: bool = False
    dry_run: bool = False

    model_config = ConfigDict(frozen=True)

def generate_build_plan(
        build_config: BuilderConfig, variants: list[str]) -> list[str]:
    build_plan = []

    if build_config.only_default_variant:
        variants = [build_config.default_variant]

    for variant in variants:
        command = 'docker image build'

        command += ' -t '
        if build_config.tag_namespace:
            command += f'{build_config.tag_namespace}/'
        command += build_config.tag_repository
        command += f':{build_config.version}-{variant}'

        # TODO: prevent setting as latest if version < latest git tag
        if (variant == build_config.default_variant
                and not build_config.tag_omit_latest):
            command += ' -t '
            if build_config.tag_namespace:
                command += f'{build_config.tag_namespace}/'
            command += build_config.tag_repository

        if build_config.build_args:
            for build_arg_key, build_arg_value in (
                    build_config.build_args.items()):
                command += f' --build-arg {build_arg_key}="{build_arg_value}"'

        command += f' variants/{variant}/.'

        build_plan.append(command)

    return build_plan


def get_builder_config(
        project_root_path: Path,
        config_file: ConfigParser,
        args: Namespace) -> BuilderConfig:
    config = {
        'tag_repository': project_root_path.stem
    }

    for section in config_file.sections():
        if section != 'build-args':
            config.update(config_file[section])

    if config_file.has_section('build-args'):
        config['build_args'] = config_file['build-args']

    informed_args = {key: value for key, value in vars(args).items() if value}

    config.update(informed_args)

    return BuilderConfig(**config)


def build(args: Namespace) -> None:
    print('doimbu - Docker Image Builder\n')

    variants = get_variants_list(EXECUTION_DIR)
    if not variants:
        print('no variants dir found')
        print('create a variant dir following the pattern:')
        print('  [CURRENT_DIR]/variants/[YOUR_VARIANT]/Dockerfile')
        return None

    config_file = get_file_config(BUILDER_CONFIG_PATH, variants)

    build_config = get_builder_config(EXECUTION_DIR, config_file, args)

    build_plan = generate_build_plan(build_config, variants)

    print('build plan:')
    for command in build_plan:
        print(command)

    if args.dry_run:
        print('\ndry run - skipping execution')
    else:
        print('\nexecuting builds..')
        for command in build_plan:
            subprocess.run(command, shell=True)
        print('builds completed')



import subprocess
from pathlib import Path

import pytest


from tests.paths import TEST_TEMP_DIR


def clean_path(path: Path) -> None:
    if path.is_file():
        path.unlink()
    elif path.is_dir():
        for dir_content_path in path.iterdir():
            clean_path(dir_content_path)
        path.rmdir()

@pytest.fixture(scope='module')
def path_test_dir() -> Path:
    return TEST_TEMP_DIR


@pytest.fixture(scope='module')
def cleanup_test_dir(path_test_dir: Path) -> None:
    clean_path(path_test_dir)

    path_test_dir.mkdir()



from __future__ import annotations

from pydantic import BaseModel, Field, PositiveInt, field_validator

DOCKER_IMAGE_FIRST_VERSION = '0.1.0'


def parse_version(version: str) -> tuple[int, int, int]:
    version_sections = version.split('.')

    for _ in range(len(version_sections), 3):
        version_sections.append('0')

    major, minor, patch = version_sections

    return int(major), int(minor), int(patch)

class Version(BaseModel):
    version: str
    major: int = Field(gt=0, default=0)
    minor: PositiveInt = 0
    patch: int = Field(gt=0, default=0)

    def model_post_init(self, *args, **kwargs):
        self.major, self.minor, self.patch = parse_version(self.version)
        self.version = f'{self.major}.{self.minor}.{self.patch}'

    # noinspection PyNestedDecorators
    @field_validator('version')
    @classmethod
    def max_3_sections(cls, value: str) -> str:
        assert len(value.split('.')) <= 3, (
            f'Invalid version {value}. '
            f'It must have at most 3 dot separated sections '
            f'(major.minor.patch).')

        return value

    # noinspection PyNestedDecorators
    @field_validator('version')
    @classmethod
    def only_positive_sections(cls, value: str) -> str:
        for section in value.split('.'):
            try:
                if int(section) < 0:
                    raise ValueError(
                        f'Version section must be greater than zero. '
                        f'{value} was found. ')
            finally:
                pass

        return value

    def __ge__(self, other_version: Version) -> bool:
        if self.major > other_version.major:
            return True
        elif self.major < other_version.major:
            return False

        if self.minor > other_version.minor:
            return True
        elif self.minor < other_version.minor:
            return False

        if self.patch > other_version.patch:
            return True
        elif self.patch < other_version.patch:
            return False

        return True

    def __repr__(self):
        return self.version



import subprocess
from configparser import ConfigParser
from pathlib import Path

import pytest
from _pytest.fixtures import SubRequest
from pytest_mock.plugin import MockerFixture

from doimbu.paths import BUILDER_CONFIG_FILE
from doimbu.sources import file_config

from tests.paths import TEST_BUILDER_CONFIG_PATH, TEST_TEMP_DIR


@pytest.mark.usefixtures("cleanup_test_dir")
@pytest.mark.parametrize(
    'default_variant,tag_section_args,build_args,'
    'execution_mode_args,git_list_tag_cmd_return',
    [
        pytest.param(
            'nobel',  # default_variant
            {},  # tag_section_args
            {},  # build_args
            {},  # execution_mode_args
            b'1.0.1\n2.2.3\n3.1\nsomething_else\n',  # git_list_tag_cmd_return
            id='only required'),
        pytest.param(
            'ubi8',  # default_variant
            {},  # tag_section_args
            {
                'linux_username': 'joe',
                'git_username': '"joe average"',
                'git_user_email': 'average.joe@pirilimpimpim.org'
            },  # build_args
            {},  # execution_mode_args
            b'1.0.1\n2.2.3\n3.1\nsomething_else\n',  # git_list_tag_cmd_return
            id='build args'),
        pytest.param(
            'nobel',  # default_variant
            {
                'tag_namespace': 'redhat',
                'tag_repository': 'ubi9'
            },  # tag_section_args
            {},  # build_args
            {},  # execution_mode_args
            b'1.0.1\n2.2.3\n3.1\nsomething_else\n',  # git_list_tag_cmd_return
            id='tag section'),
        pytest.param(
            'nobel',  # default_variant
            {},  # tag_section_args
            {},  # build_args
            {
                'only_default_variant': True,
                'tag_omit_latest': True,
                'dry_run': True
            },  # execution_mode_args
            b'1.0.1\n2.2.3\n3.1\nsomething_else\n',  # git_list_tag_cmd_return
            id='execution-mode section'),
        pytest.param(
            'nobel',  # default_variant
            {'tag_namespace': 'redhat'},  # tag_section_args
            {'linux_username': 'joe'},  # build_args
            {'dry_run': True},  # execution_mode_args
            b'1.0.1\n2.2.3\n3.1\nsomething_else\n',  # git_list_tag_cmd_return
            id='multiple section')
    ]
)
def test_write_file_config(
        default_variant: str,
        tag_section_args: dict[str, str],
        build_args: dict[str, str] | None,
        execution_mode_args: dict[str, bool],
        git_list_tag_cmd_return: bytes,
        mocker: MockerFixture) -> None:
    mocker.patch(
        'doimbu.sources.git.subprocess.run',
        return_value=subprocess.CompletedProcess(
            [], 0, stdout=git_list_tag_cmd_return))

    file_config.write_file_config(
        config_path=TEST_BUILDER_CONFIG_PATH,
        default_variant=default_variant,
        build_args=build_args,
        **tag_section_args,
        **execution_mode_args)

    config = ConfigParser()
    config.read(TEST_BUILDER_CONFIG_PATH)

    assert 'doimbu' in config.sections()
    assert config['doimbu']['default_variant'] == default_variant

    config_sections_args_map = {
        'tag': tag_section_args,
        'build-args': build_args
    }
    for config_section, config_items in config_sections_args_map.items():
        if config_items:
            for config_item_key, config_item_value in config_items.items():
                assert (config[config_section][config_item_key]
                        == config_item_value)

    for key, value in execution_mode_args.items():
        assert config['execution-mode'].getboolean(key) == value


def generate_config_from_dict(sections: list[dict[str, any]]) -> ConfigParser:
    config = ConfigParser()

    for section in sections:
        config[section['section_name']] = section['section_items']

    return config


@pytest.fixture(
    params=[
        {
            'scenario_dir': 'ok_file',
            'file_config': [
                {
                    'section_name': 'doimbu',
                    'section_items': {
                        'default_variant': 'nobel'
                    }
                },
                {
                    'section_name': 'tag',
                    'section_items': {
                        'version': '1.2.0'
                    }
                }
            ],
            'git_list_tag_cmd_return': b'1.0.1\n2.2.3\n3.1\nsomething_else\n',
            'expected_config': [
                {
                    'section_name': 'doimbu',
                    'section_items': {
                        'default_variant': 'nobel'
                    }
                },
                {
                    'section_name': 'tag',
                    'section_items': {
                        'version': '1.2.0'
                    }
                }
            ]
        },
        {
            'scenario_dir': 'file_missing_mandatory_section',
            'file_config': [
                {
                    'section_name': 'tag',
                    'section_items': {
                        'version': '1.2.0',
                        'tag_namespace': 'redhat'
                    }
                }
            ],
            'git_list_tag_cmd_return': b'1.0.1\n2.2.3\n3.1\nsomething_else\n',
            'expected_config': [
                {
                    'section_name': 'doimbu',
                    'section_items': {
                        'default_variant': 'ubi'
                    }
                },
                {
                    'section_name': 'tag',
                    'section_items': {
                        'version': '3.1.0',
                        'tag_namespace': 'redhat'
                    }
                }
            ]
        },
        {
            'scenario_dir': 'file_missing_mandatory_item',
            'file_config': [
                {
                    'section_name': 'doimbu',
                    'section_items': {
                        'something': 'x'
                    }
                }
            ],
            'git_list_tag_cmd_return': b'1.0.1\n2.2.3\n3.1\nsomething_else\n',
            'expected_config': [
                {
                    'section_name': 'doimbu',
                    'section_items': {
                        'something': 'x',
                        'default_variant': 'ubi'

                    }
                },
                {
                    'section_name': 'tag',
                    'section_items': {
                        'version': '3.1.0'
                    }
                }
            ]
        },
        {
            'scenario_dir': 'missing_config_file',
            'file_config': None,
            'git_list_tag_cmd_return': b'1.0.1\n2.2.3\n3.1\nsomething_else\n',
            'expected_config': [
                {
                    'section_name': 'doimbu',
                    'section_items': {
                        'default_variant': 'ubi'
                    }
                },
                {
                    'section_name': 'tag',
                    'section_items': {
                        'version': '3.1.0'
                    }
                }
            ]
        },
    ],
    ids=[
        'config file ok',
        'config file missing mandatory section',
        'config file missing mandatory item',
        'missing config file'
    ]
)
def params_test_get_file_config(
        request: SubRequest,
        mocker: MockerFixture) -> dict[str, Path | ConfigParser]:

    base_test_dir = TEST_TEMP_DIR / 'test_get_file_config'
    test_config_dir = base_test_dir / request.param['scenario_dir']
    test_config_path = test_config_dir / BUILDER_CONFIG_FILE

    test_config_dir.mkdir(parents=True)

    if request.param['file_config']:
        config = generate_config_from_dict(request.param['file_config'])

        with test_config_path.open('w') as f:
            config.write(f)

    mocker.patch(
        'doimbu.sources.git.subprocess.run',
        return_value=subprocess.CompletedProcess(
            [], 0, stdout=request.param['git_list_tag_cmd_return']))

    expected_config = generate_config_from_dict(
        request.param['expected_config'])

    return {
        'config_path': test_config_path,
        'variants': ['ubi', 'nobel', 'bullseye'],
        'expected_file_config': expected_config
    }


@pytest.mark.usefixtures("cleanup_test_dir")
def test_get_file_config(
        params_test_get_file_config: dict[str, Path | ConfigParser]) -> None:
    config_file = file_config.get_file_config(
        params_test_get_file_config['config_path'],
        params_test_get_file_config['variants']
    )

    expected_config = params_test_get_file_config['expected_file_config']

    assert set(config_file.sections()) == set(expected_config.sections())

    for section in expected_config.sections():
        assert config_file.items(section) == expected_config.items(section)



from pathlib import Path


EXECUTION_DIR = Path().cwd()

BUILDER_CONFIG_FILE = 'doimbu.ini'
BUILDER_CONFIG_PATH = EXECUTION_DIR / BUILDER_CONFIG_FILE
DOCKER_IMAGE_VARIANTS_DIR = EXECUTION_DIR / 'variants'
LOG_PATH = EXECUTION_DIR / 'build_output.log'



from pathlib import Path

from doimbu.sources import dir_structure

import pytest
from _pytest.fixtures import SubRequest


@pytest.fixture(
    params=[
        {
            'scenario_dir': 'no_variants_dir',
            'variants_dir': None,
            'variants': []
        },
        {
            'scenario_dir': 'empty_variants_dir',
            'variants_dir': 'variants',
            'variants': []
        },
        {
            'scenario_dir': 'single_variant',
            'variants_dir': 'variants',
            'variants': ['nobel']
        },
        {
            'scenario_dir': 'multiple_variants',
            'variants_dir': 'variants',
            'variants': ['nobel', 'ubi8', 'bookworm']
        }
    ],
    ids=[
        'no variants dir',
        'empty variants dir',
        'single variant',
        'multiple variant'
    ]
)
def prepare_variants_test_dir(
        path_test_dir: Path,
        request: SubRequest) -> dict[str, Path | list[str]]:
    scenario_dir = path_test_dir / request.param['scenario_dir']
    scenario_dir.mkdir()

    if request.param['variants_dir']:
        variants_dir = scenario_dir / request.param['variants_dir']
        variants_dir.mkdir()

        for variant in request.param['variants']:
            variant_dir = variants_dir / variant
            variant_dir.mkdir()

    return {
        'scenario_dir': scenario_dir,
        'expected_variants': request.param['variants']
    }


@pytest.mark.usefixtures("cleanup_test_dir")
def test_get_variants_list(prepare_variants_test_dir: dict[str, any]) -> None:
    variants = dir_structure.get_variants_list(
        prepare_variants_test_dir['scenario_dir'])

    assert sorted(variants) == sorted(
        prepare_variants_test_dir['expected_variants'])


from configparser import ConfigParser
from pathlib import Path

from doimbu.version import Version
from doimbu.sources.git import get_latest_git_version_tag


def create_file_config(
        default_variant: str,
        version: Version | None = None,
        tag_namespace: str | None = None,
        tag_repository: str | None = None,
        build_args: dict[str, str] | None = None,
        only_default_variant: bool = False,
        tag_omit_latest: bool = False,
        dry_run: bool = False) -> ConfigParser:
    """Create the configurations that will be saved on the file system.

    It's a different concept from the BuilderConfig class.
    The BuilderConfig class represents the runtime configuration, that merges
    the configuration present on the file config(created by this function),
    command line parameters, git tags and file system directory structure
    (variants).
    """
    config = ConfigParser()

    _add_mandatory_config_items(
        config, default_variant=default_variant, version=version)

    tag_section = {
        'tag_namespace': tag_namespace,
        'tag_repository': tag_repository
    }
    tag_section = {key: value for key, value in tag_section.items()
                   if value is not None}
    if tag_section:
        config['tag'].update(tag_section)

    if build_args is not None:
        config['build-args'] = build_args

    if any({only_default_variant, tag_omit_latest, dry_run}):
        execution_mode_section = {
            'only_default_variant': only_default_variant,
            'tag_omit_latest': tag_omit_latest,
            'dry_run': dry_run
        }
        execution_mode_section = {
            key: str(value) for key, value in execution_mode_section.items()
            if value is not None
        }
        if execution_mode_section:
            config['execution-mode'] = execution_mode_section

    return config


def _has_mandatory_config_items(config: ConfigParser) -> bool:
    has_default_variant = False
    if config.has_section('doimbu'):
        has_default_variant = (
                config['doimbu'].get('default_variant') is not None)

    has_tag_version = False
    if config.has_section('tag'):
        has_tag_version = config['tag'].get('version') is not None

    return all({has_default_variant, has_tag_version})


def _add_mandatory_config_items(
        config: ConfigParser,
        default_variant: str,
        version: Version | None = None) -> None:
    if not config.has_section('doimbu'):
        config.add_section('doimbu')

    config['doimbu'].update({'default_variant': default_variant})

    if not config.has_section('tag'):
        config.add_section('tag')

    version = version if version else get_latest_git_version_tag()

    config['tag'].update({'version': version.version})


def get_file_config(config_path: Path, variants: list[str]) -> ConfigParser:
    if not config_path.is_file():
        write_file_config(config_path, default_variant=variants[0])

    config = ConfigParser()
    config.read(config_path)

    if not _has_mandatory_config_items(config):
        _add_mandatory_config_items(config, default_variant=variants[0])

    return config


def write_file_config(
        config_path: Path,
        default_variant: str,
        version: Version | None = None,
        tag_namespace: str | None = None,
        tag_repository: str | None = None,
        build_args: dict[str, str] | None = None,
        only_default_variant: bool = False,
        tag_omit_latest: bool = False,
        dry_run: bool = False) -> None:
    config = create_file_config(
        default_variant,
        version,
        tag_namespace,
        tag_repository,
        build_args,
        only_default_variant,
        tag_omit_latest,
        dry_run)

    with config_path.open('w') as f:
        config.write(f)



from argparse import Namespace
from configparser import ConfigParser
from pathlib import Path

import pytest

from doimbu import builder
from doimbu.builder import BuilderConfig
from doimbu.sources import file_config
from doimbu.version import Version


@pytest.mark.parametrize(
    'build_config,variants,expected_build_plan',
    [
        pytest.param(
            BuilderConfig(
                version='1.1.0',
                default_variant='nobel',
                tag_repository='mongodb-community-server'),  # build_config
            ['nobel'],  # variants
            [
                'docker image build -t mongodb-community-server:1.1.0-nobel'
                ' -t mongodb-community-server variants/nobel/.'
            ],  # expected_build_plan
            id='single variant | minimal BuildConfig'
        ),
        pytest.param(
            BuilderConfig(
                version='1.1.0',
                default_variant='nobel',
                tag_repository='mongodb-community-server'),  # build_config
            ['bullseye', 'nobel', 'ubi8'],  # variants
            [
                'docker image build -t mongodb-community-server:1.1.0-bullseye'
                ' variants/bullseye/.',
                'docker image build -t mongodb-community-server:1.1.0-nobel'
                ' -t mongodb-community-server variants/nobel/.',
                'docker image build -t mongodb-community-server:1.1.0-ubi8'
                ' variants/ubi8/.'
            ],  # expected_build_plan
            id='3 variants | minimal BuildConfig'
        ),
        pytest.param(
            BuilderConfig(
                version='1.1.0',
                default_variant='nobel',
                tag_namespace='mongodb',
                tag_repository='mongodb-community-server',
                build_args={
                    'git_username': 'some git user',
                    'git_user_email': 'someone@somecompany.com'
                },
                only_default_variant=True,
                tag_omit_latest=True,
                dry_run=True),  # build_config
            ['bullseye', 'nobel'],  # variants
            [
                'docker image build -t mongodb/mongodb-community-server'
                ':1.1.0-nobel --build-arg git_username="some git user"'
                ' --build-arg git_user_email="someone@somecompany.com"'
                ' variants/nobel/.'
            ],  # expected_build_plan
            id='2 variants | complete BuildConfig'
        )
    ]
)
def test_generate_build_plan(
        build_config: builder.BuilderConfig,
        variants: list[str],
        expected_build_plan: list[str]) -> None:
    assert (builder.generate_build_plan(build_config, variants)
            == expected_build_plan)


@pytest.mark.parametrize(
    'project_root_path,config_file,args,expected_builder_config',
    [
        pytest.param(
            Path('/path/to/image/bed_machine'),  # project_root_path
            file_config.create_file_config(
                'ubi8',
                Version(version='3.1.0'),
                'mongodb',
                'mongodb-community-server',
                {
                    'username': 'thomas thompson',
                    'user_email': 'thomas.thompson@mongodb.com',
                },
                True,
                True,
                True),  # config_file
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # args
            builder.BuilderConfig(
                version='3.1.0',
                default_variant='ubi8',
                tag_namespace='mongodb',
                tag_repository='mongodb-community-server',
                build_args={
                    'username': 'thomas thompson',
                    'user_email': 'thomas.thompson@mongodb.com'
                },
                only_default_variant=True,
                tag_omit_latest=True,
                dry_run=True),  # expected_builder_config
            id='no params / full config file'),
        pytest.param(
            Path('/path/to/image/bed_machine'),  # project_root_path
            file_config.create_file_config(
                'ubi8',
                version=Version(version='3.1.0')),  # config_file
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # args
            builder.BuilderConfig(
                version='3.1.0',
                default_variant='ubi8',
                tag_repository='bed_machine'),  # expected_builder_config
            id='no params / only mandatory config file'),
        pytest.param(
            Path('/path/to/image/bed_machine'),  # project_root_path
            file_config.create_file_config(
                'ubi8',
                version=Version(version='3.1.0'),
                only_default_variant=True),  # config_file
            Namespace(
                default_variant='nobel',
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # args  # args
            builder.BuilderConfig(
                version='3.1.0',
                default_variant='nobel',
                tag_repository='bed_machine',
                only_default_variant=True),  # expected_builder_config
            id='partial params / partial config file'),
        pytest.param(
            Path('/path/to/image/bed_machine'),  # project_root_path
            file_config.create_file_config(
                'ubi8',
                version=Version(version='3.1.0'),
                tag_namespace='mongodb',
                tag_repository='mongodb-community-server'),  # config_file
            Namespace(
                default_variant='nobel',
                tag_namespace='mongodb',
                tag_repository='mongodb-enterprise-server',
                build_args={
                    'linux_username': 'some_linux_username',
                },
                only_default_variant=True,
                tag_omit_latest=True,
                dry_run=True),  # args
            builder.BuilderConfig(
                version='3.1.0',
                default_variant='nobel',
                tag_namespace='mongodb',
                tag_repository='mongodb-enterprise-server',
                build_args={
                    'linux_username': 'some_linux_username',
                },
                only_default_variant=True,
                tag_omit_latest=True,
                dry_run=True
            ),  # expected_builder_config
            id='all params / all config file')
    ]
)
def test_get_builder_config(
        project_root_path: Path,
        config_file: ConfigParser,
        args: Namespace,
        expected_builder_config: builder.BuilderConfig | None) -> None:

    if expected_builder_config:
        assert (builder.get_builder_config(
            project_root_path, config_file, args) == expected_builder_config)
    else:
        with pytest.raises(KeyError, match=r"doimbu"):
            builder.get_builder_config(project_root_path, config_file, args)



import subprocess

from pydantic import ValidationError

from doimbu.version import DOCKER_IMAGE_FIRST_VERSION, Version


GIT_FETCH_AND_LIST_TAGS_CMD = (
    'git fetch --tags & git tag --sort=version:refname')


def get_latest_git_version_tag() -> Version:
    """Get the latest git version tag.

    - Fetches all tags from remote
    - Retrieve an ordered list of tags from git that contains version
    and non-version tags.

    :return: return the latest tag that matches the version pattern
        (major.minor.patch - where minor and patch are optional)
    """
    completed_process = subprocess.run(
        GIT_FETCH_AND_LIST_TAGS_CMD, shell=True, capture_output=True)

    stdout = completed_process.stdout.decode()

    if stdout:
        versions = stdout.split('\n')

        for version in reversed(versions):
            try:
                return Version(version=version)
            except ValidationError:
                pass

    return Version(version=DOCKER_IMAGE_FIRST_VERSION)


from contextlib import nullcontext as does_not_raise

import pytest
from _pytest.python_api import RaisesContext
from pydantic_core import ValidationError

from doimbu.version import parse_version, Version


@pytest.mark.parametrize(
    'version,expected_major,expected_minor,expected_patch',
    [
        pytest.param('1', 1,  0, 0, id='only major'),
        pytest.param('0.1', 0, 1, 0, id='major_minor'),
        pytest.param('3.2.1', 3, 2, 1, id='major_minor_patch'),
    ]
)
def test_parse_version(
        version: str,
        expected_major: int,
        expected_minor: int,
        expected_patch: int) -> None:
    assert parse_version(version) == (
        expected_major, expected_minor, expected_patch
    )


@pytest.mark.parametrize(
    'current_version,other_version,expected_ge_result',
    [
        pytest.param(
            Version(version='1.1.1'),
            Version(version='1.1.1'),
            True,
            id='equal'),
        pytest.param(
            Version(version='2.0.1'),
            Version(version='1.2.1'),
            True,
            id='bigger major'),
        pytest.param(
            Version(version='2.0.1'),
            Version(version='3.2.1'),
            False,
            id='smaller major'),
        pytest.param(
            Version(version='2.10.1'),
            Version(version='2.2.1'),
            True,
            id='bigger minor'),
        pytest.param(
            Version(version='2.0.1',),
            Version(version='2.2.1'),
            False,
            id='smaller minor'),
        pytest.param(
            Version(version='3.0.11'),
            Version(version='3.0.1'),
            True,
            id='bigger patch'),
        pytest.param(
            Version(version='3.0.1'),
            Version(version='3.0.2'),
            False,
            id='smaller patch')
    ])
def test_version_ge(
        current_version: Version,
        other_version: Version,
        expected_ge_result: bool) -> None:
    assert (current_version >= other_version) == expected_ge_result


@pytest.mark.parametrize(
    'version,expected_version,expected_exception',
    [
        pytest.param(
            '3.2.1', # version
            Version(version='3.2.1'),  # expected_version
            does_not_raise(),  # expected_exception
            id='3.2.1 - major minor patch'),
        pytest.param(
            '1.0.0',  # version
            Version(version='1.0.0'),  # expected_version
            does_not_raise(),  # expected_exception
            id='1.0.0 - major'),
        pytest.param(
            '1',  # version
            Version(version='1.0.0'),  # expected_version
            does_not_raise(),  # expected_exception
            id='1 - major compact'),
        pytest.param(
            '0.2.0',
            Version(version='0.2.0'),  # expected_version
            does_not_raise(),  # expected_exception
            id='0.2.0 - minor'),
        pytest.param(
            '0.2',  # version
            Version(version='0.2.0'),  # expected_version
            does_not_raise(),  # expected_exception
            id='0.2 - minor compact'),
        pytest.param(
            '1.2.3.4',  # version
            None,  # expected_version
            pytest.raises(ValidationError),  # expected_exception
            id='1.2.3.4 - too many sections'),
        pytest.param(
            '0.-3.0',  # version
            None,  # expected_version
            pytest.raises(ValidationError),  # expected_exception
            id='0.-3.0 - negative section'),
        pytest.param(
            '1.2.x',  # version
            None,  # expected_version
            pytest.raises(ValidationError),  # expected_exception
            id='1.2.x - not int section')
    ]
)
def test_version_validation(
        version: str,
        expected_version: Version | None,
        expected_exception: does_not_raise | RaisesContext) -> None:
    with expected_exception:
        assert Version(version=version) == expected_version


from pathlib import Path
from doimbu.paths import BUILDER_CONFIG_FILE


ROOT_DIR = Path(__file__).parent.parent
TEST_ROOT_DIR = ROOT_DIR / 'tests'
TEST_TEMP_DIR = TEST_ROOT_DIR / 'temp'
TEST_BUILDER_CONFIG_PATH = TEST_TEMP_DIR / BUILDER_CONFIG_FILE



import pytest
from argparse import Namespace
from pytest_mock.plugin import MockerFixture

from doimbu.__main__ import parse_args


@pytest.mark.parametrize(
    'args,expected_namespace',
    [
        pytest.param(
            [],  # args
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # expected_namespace (parsed args)
            id='No args'),
        pytest.param(
            ['-d', 'nobel'],  # args
            Namespace(
                default_variant='nobel',
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # expected_namespace (parsed args)
            id='default-variant short'),
        pytest.param(
            ['--default-variant', 'nobel'],  # args
            Namespace(
                default_variant='nobel',
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # expected_namespace (parsed args)
            id='default-variant long'),
        pytest.param(
            ['--tag-namespace', 'some-org'],
            Namespace(
                default_variant=None,
                tag_namespace='some-org',
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),
            id='tag-namespace'),
        pytest.param(
            ['--tag-repository', 'some-repo'],
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository='some-repo',
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),
            id='tag-repository'),
        pytest.param(
            ['-d', 'ubi8', '-n', 'other-org', '-r', 'other-repo'],
            Namespace(
                default_variant='ubi8',
                tag_namespace='other-org',
                tag_repository='other-repo',
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),
            id='multiple params'),
        pytest.param(
            ['--build-arg', 'linux_username=some_linux_username'],
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args={
                    'linux_username': 'some_linux_username',
                },
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),
            id='build-arg'),
        pytest.param(
            [
                '--build-arg', 'linux_username=some_linux_username',
                '--build-arg', 'git_user_email=someone@somecompany.com'
            ],
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args={
                    'linux_username': 'some_linux_username',
                    'git_user_email': 'someone@somecompany.com'
                },
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),
            id='multiple build-args'),
        pytest.param(
            ['--build-arg', 'git_username="some git user"'],
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args={
                    'git_username': 'some git user',
                },
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),
            id='build-arg with space'),

        pytest.param(
            ['--only-default-variant'],
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=True,
                tag_omit_latest=False,
                dry_run=False),
            id='only-default-variant'),
        pytest.param(
            ['--tag-omit-latest'], Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=True,
                dry_run=False),
            id='tag-omit-latest'),
        pytest.param(
            ['--dry-run'],
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version=None,
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=True),
            id='dry-run'),
        pytest.param(
            ['--version', '0.7.0'],  # args
            Namespace(
                default_variant=None,
                tag_namespace=None,
                tag_repository=None,
                build_args=None,
                version='0.7.0',
                only_default_variant=False,
                tag_omit_latest=False,
                dry_run=False),  # expected_namespace (parsed args)
            id='version')
    ])
def test_parse_args(
        mocker: MockerFixture,
        args: list[str],
        expected_namespace: Namespace) -> None:
    if not args:
        mocker.patch(
            'doimbu.__main__.argparse._sys.argv',
            return_value=['doimbu.__main__.py'])

    assert parse_args(*args) == expected_namespace


import argparse

from doimbu import builder


def _copy_dict(dictionary: dict):
    return dictionary.copy() if dictionary else {}

class AppendKeyValueAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        dictionary = getattr(namespace, self.dest)
        dictionary = _copy_dict(dictionary)
        key, value = values.split('=')
        dictionary[key] = value.replace('"', '')
        setattr(namespace, self.dest, dictionary)


def parse_args(*args: str) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description='doimbu - Docker Image Builder')
    parser.add_argument(
        '-d', '--default-variant',
        help='set which variant is the default'
             ' (the one that will receive the additional tag suffix `latest`)')
    parser.add_argument(
        '-n', '--tag-namespace',
        help='[tag-namespace] for the tag pattern:'
             ' [tag-namespace]/[tag-repository]:[version]-[variant]'
             ' | example: mongodb/mongodb-community-server:1.2.1-ubi8'
    )
    parser.add_argument(
        '-r', '--tag-repository',
        help='[tag-repository] for the tag pattern:'
             ' [tag-namespace]/[tag-repository]:[version]-[variant]'
             ' | example: mongodb/mongodb-community-server:1.2.1-ubi8'
    )
    parser.add_argument(
        '-v', '--version',
        help='[version] for the tag pattern:'
             ' [tag-namespace]/[tag-repository]:[version]-[variant]'
             ' | example: mongodb/mongodb-community-server:1.2.1-ubi8'
    ),
    parser.register('action', 'append_key_value', AppendKeyValueAction)
    parser.add_argument(
        '-b', '--build-arg',
        dest='build_args',
        action='append_key_value',
        help='same as `docker image build --build-arg`'
             ' | can be used multiple times'
             ' | enclosed by \''
             ' | example: --build-arg \'git_username="jhonny joe"\''
    )
    parser.add_argument(
        '-o', '--only-default-variant',
        action='store_true',
        help='build only the default variant')
    parser.add_argument(
        '-l', '--tag-omit-latest',
        action='store_true',
        help='do not create at tag with the `latest` suffix'
             ' for the default variant')
    parser.add_argument(
        '-y', '--dry-run',
        action='store_true',
        help='print the commands on the execution plan'
             ', but do not execute them')

    return parser.parse_args(args) if args else parser.parse_args()


if __name__ == '__main__':
    _args = parse_args()

    builder.build(_args)
